import 'dart:async';
import 'dart:collection';

import 'package:flutter/widgets.dart';

import 'rx_interface.dart';
import 'rx_stream.dart';

mixin RxMixin<T> {
  final _key = UniqueKey();
  final _rxStream = RxStream<T>();
  final _subPool = <Key, StreamSubscription>{};

  bool get canUpdate => _subPool.isNotEmpty;

  LightSubscription<T> listen(
    void Function(T event) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError = false,
  }) =>
      _rxStream.listen(onData, onError: onError, onDone: onDone, cancelOnError: cancelOnError);

  void addListener(Key key, RxStream<T> rxStream) {
    if (_subPool.containsKey(key)) return;
    _subPool[key] = rxStream.listen((event) => _rxStream.addData(event));
  }

  void close() {
    _subPool.values.forEach((item) => item.cancel());
    _subPool.clear();
    _rxStream.close();
  }
}

mixin RxValueMixin<T> on RxMixin<T> {
  bool _firstRebuild = true;

  T _value;

  // T call([T v]) {
  //   if (v != null) {
  //     value = v;
  //   }
  //   return value;
  // }

  T get value {
    if (RxInterface.proxy != null) {
      RxInterface.proxy.addListener(_key, _rxStream);
    }
    return _value;
  }

  set value(T value) {
    // 第一次更新时,即便值相等也会更新
    if (_value == value && !_firstRebuild) return;
    _firstRebuild = false;
    _value = value;
    _rxStream.addData(_value);
  }

  void refresh() => _rxStream.addData(value);

  void nil() => _rxStream.addData(_value = null);

  void addError(Object error, [StackTrace stackTrace]) => _rxStream.addError(error, stackTrace);

  @override
  int get hashCode => value.hashCode;

  @override
  String toString() => value.toString();

  @override
  bool operator ==(Object other) {
    if (other is T) return other == value;
    if (other is RxValueMixin<T>) return other.value == value;
    return false;
  }
}

mixin RxObjectMixin<T> on RxValueMixin<T> {
  void update(void func(T obj)) {
    func(_value);
    _rxStream.addData(_value);
  }
}

class RxNotifier<T> = RxInterface<T> with RxMixin<T>;

abstract class RxImpl<T> extends RxNotifier<T> with RxValueMixin<T> {
  RxImpl([T initial]) {
    _value = initial;
  }
}

abstract class RxSetNotifier<T> = SetMixin<T> with RxMixin<Set<T>>;

abstract class RxSetImpl<T> extends RxSetNotifier<T> with RxValueMixin<Set<T>> implements RxInterface<Set<T>> {
  RxSetImpl([Set<T> initial = const {}]) {
    if (initial != null) _value = Set.from(initial);
  }
}

abstract class RxMapNotifier<K, V> = MapMixin<K, V> with RxMixin<Map<K, V>>;

abstract class RxMapImp<K, V> extends RxMapNotifier<K, V> with RxValueMixin<Map<K, V>> implements RxInterface<Map<K, V>> {
  RxMapImp([Map<K, V> initial = const {}]) {
    if (initial != null) _value = Map.from(initial);
  }
}

abstract class RxListNotifier<T> = ListMixin<T> with RxMixin<List<T>>;

abstract class RxListImp<T> extends RxListNotifier<T> with RxValueMixin<List<T>> implements RxInterface<List<T>> {
  RxListImp([List<T> initial = const []]) {
    if (initial != null) _value = List.from(initial);
  }
}
