import 'dart:async';

class RxStream<T> {
  final void Function() onPause;
  final void Function() onResume;
  final FutureOr<void> Function() onCancel;

  RxStream({
    this.onPause,
    this.onResume,
    this.onCancel,
  });

  List<LightSubscription<T>> _subList = [];

  int get length => _subList?.length;

  bool get isClosed => _subList == null;

  bool _isBusy = false;

  LightSubscription<T> listen(
    void Function(T event) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    var subscription = LightSubscription<T>(
      _removeSubscription,
      onPause: onPause,
      onResume: onResume,
      onCancel: onCancel,
    )
      ..onData(onData)
      ..onError(onError)
      ..onDone(onDone)
      ..cancelOnError = cancelOnError;
    _addSubscription(subscription);
    return subscription;
  }

  FutureOr<void> _addSubscription(LightSubscription<T> subscription) async {
    if (_isBusy) {
      await Future.delayed(Duration.zero);
    }
    return _subList.add(subscription);
  }

  FutureOr<bool> _removeSubscription(LightSubscription<T> subscription) async {
    if (_isBusy) {
      await Future.delayed(Duration.zero);
    }
    return _subList.remove(subscription);
  }

  void addData(T event) {
    assert(!isClosed, '无法添加event到一个已关闭的Stream.');
    _notifyData(event);
  }

  void _notifyData(T data) {
    _isBusy = true;
    _subList.where((item) => !item.isPaused).forEach((item) => item.data?.call(data));
    _isBusy = false;
  }

  void addError(Object error, [StackTrace stackTrace]) {
    assert(!isClosed, '无法添加error到一个已关闭的Stream.');
    _notifyError(error, stackTrace);
  }

  void _notifyError(Object error, [StackTrace stackTrace]) {
    _isBusy = true;
    var waitToRemove = <LightSubscription<T>>[];
    _subList.where((item) => !item.isPaused).forEach((item) {
      if (stackTrace == null) {
        item.error?.call(error);
      } else {
        item.error?.call(error, stackTrace);
      }
      if (item.cancelOnError) {
        waitToRemove.add(item);
        item.pause();
        item.done?.call();
      }
    });
    waitToRemove.forEach((item) => _subList.remove(item));
    _isBusy = false;
  }

  void close() {
    assert(!isClosed, '无法关闭一个已关闭的Stream.');
    _notifyDone();
    _subList = null;
  }

  void _notifyDone() {
    _isBusy = true;
    _subList.where((item) => !item.isPaused).forEach((item) => item.done?.call());
    _isBusy = false;
  }
}

class LightSubscription<T> extends StreamSubscription<T> {
  final FutureOr<bool> Function(LightSubscription<T> subscription) removeSubscription;
  final Function() onPause;
  final Function() onResume;
  final FutureOr<void> Function() onCancel;

  LightSubscription(this.removeSubscription, {this.onPause, this.onResume, this.onCancel});

  Function(T value) data;
  Function error;
  Function() done;
  bool cancelOnError = false;

  bool _isPaused = false;

  @override
  void onData(Function(T value) handleData) => data = handleData;

  @override
  void onError(Function handleError) => error = handleError;

  @override
  void onDone(Function() handleDone) => done = handleDone;

  @override
  void pause([Future<void> resumeSignal]) {
    _isPaused = true;
    onPause?.call();
  }

  @override
  void resume() {
    _isPaused = false;
    onResume?.call();
  }

  @override
  Future<Function> cancel() {
    removeSubscription(this);
    onCancel?.call();
    return Future.value();
  }

  @override
  bool get isPaused => _isPaused;

  @override
  Future<T> asFuture<T>([T futureValue]) => Future.value(futureValue);
}
