import 'dart:async';

import '../core/rx_interface.dart';

typedef WorkerCallback<T> = Function(T callback);

class Worker {
  final Future<void> Function() worker;

  Worker(this.worker);

  bool _isDisposed = false;

  bool get isDisposed => _isDisposed;

  void dispose() {
    if (isDisposed) return;
    _isDisposed = true;
    worker.call();
  }
}

Worker every<T>(
  RxInterface<T> rxObj,
  WorkerCallback<T> callback, {
  bool condition(T event),
  Function onError,
  void Function() onDone,
  bool cancelOnError,
}) {
  assert(callback != null);

  var sub = rxObj.listen(
    (event) {
      if (condition == null || (condition != null && condition(event))) {
        callback.call(event);
      }
    },
    onError: onError,
    onDone: onDone,
    cancelOnError: cancelOnError,
  );
  return Worker(sub.cancel);
}

Worker once<T>(
  RxInterface<T> rxObj,
  WorkerCallback<T> callback, {
  bool condition(T event),
  Function onError,
  void Function() onDone,
  bool cancelOnError,
}) {
  assert(callback != null);

  Worker worker;
  var sub = rxObj.listen(
    (event) {
      if (condition == null || (condition != null && condition(event))) {
        worker.dispose();
        callback.call(event);
      }
    },
    onError: onError,
    onDone: onDone,
    cancelOnError: cancelOnError,
  );
  worker = Worker(sub.cancel);
  return worker;
}

Worker interval<T>(
  RxInterface<T> rxObj,
  WorkerCallback<T> callback, {
  Duration duration = const Duration(seconds: 1),
  Function onError,
  void Function() onDone,
  bool cancelOnError,
}) {
  assert(callback != null);
  assert(duration != null);

  var debounce = false;
  var sub = rxObj.listen(
    (event) async {
      if (debounce) return;
      debounce = true;
      callback.call(event);
      await Future.delayed(duration);
      debounce = false;
    },
    onError: onError,
    onDone: onDone,
    cancelOnError: cancelOnError,
  );
  return Worker(sub.cancel);
}

Worker debounce<T>(
  RxInterface<T> rxObj,
  WorkerCallback<T> callback, {
  Duration duration = const Duration(seconds: 1),
  Function onError,
  void Function() onDone,
  bool cancelOnError,
}) {
  assert(callback != null);
  assert(duration != null);

  final _debounce = _Debounce(duration);
  var sub = rxObj.listen(
    (event) {
      _debounce.call(() {
        callback.call(event);
      });
    },
    onError: onError,
    onDone: onDone,
    cancelOnError: cancelOnError,
  );
  return Worker(sub.cancel);
}

class _Debounce {
  final Duration duration;

  _Debounce(this.duration);

  Timer _timer;

  void call(void Function() action) {
    _timer?.cancel();
    _timer = Timer(duration, action);
  }

  void cancel() => _timer?.cancel();
}
