import 'package:flutter/widgets.dart';

import 'package:yuro_core/yuro_core.dart';
import 'package:yuro_instance/yuro_instance.dart';

import 'yuro_controller.dart';

typedef YuroControllerBuilder<T extends DisposableInterface> = Widget Function(T controller);

class YuroBuilder<T extends YuroController> extends StatefulWidget {
  final String id;
  final String tag;
  final T init;
  final YuroControllerBuilder<T> builder;
  final bool global;
  final bool autoRemove;
  final void Function(State state) initState, dispose, didChangeDependencies;
  final void Function(YuroBuilder oldWidget, State state) didUpdateWidget;

  const YuroBuilder({
    Key key,
    this.id,
    this.tag,
    this.init,
    @required this.builder,
    this.global = true,
    this.autoRemove = true,
    this.initState,
    this.dispose,
    this.didChangeDependencies,
    this.didUpdateWidget,
  })  : assert(builder != null),
        super(key: key);

  @override
  _YuroBuilderState<T> createState() => _YuroBuilderState<T>();
}

mixin YuroStateUpdateMixin<T extends StatefulWidget> on State<T> {
  void getUpdate() {
    if (mounted) setState(() {});
  }
}

class _YuroBuilderState<T extends YuroController> extends State<YuroBuilder<T>> with YuroStateUpdateMixin {
  T controller;

  // 是否是controller的创建者
  bool isCreator = false;
  VoidCallback remove;

  @override
  void initState() {
    super.initState();
    if (widget.global) {
      final isPrepared = Yuro.isPrepared<T>(tag: widget.tag);
      final isRegistered = Yuro.isRegistered<T>(tag: widget.tag);
      if (isPrepared) {
        if (Yuro.smartManagement == SmartManagement.keepFactory) {
          isCreator = true;
        }
        controller = Yuro.find<T>(tag: widget.tag);
      } else if (isRegistered) {
        controller = Yuro.find<T>(tag: widget.tag);
        isCreator = false;
      } else {
        controller = widget.init;
        isCreator = true;
        Yuro.put<T>(controller, tag: widget.tag);
      }
    } else {
      controller = widget.init;
      isCreator = true;
      controller?.onStart();
    }
    _subscribeToController();
    widget.initState?.call(this);
  }

  @override
  void didChangeDependencies() {
    super.didChangeDependencies();
    widget.didChangeDependencies?.call(this);
  }

  @override
  void didUpdateWidget(covariant YuroBuilder<T> oldWidget) {
    super.didUpdateWidget(oldWidget);
    if (oldWidget.id != widget.id) {
      _subscribeToController();
    }
    widget.didUpdateWidget?.call(oldWidget, this);
  }

  @override
  void dispose() {
    super.dispose();
    widget.dispose?.call(this);
    if (isCreator && widget.autoRemove) {
      if (Yuro.isRegistered<T>(tag: widget.tag)) {
        Yuro.delete<T>(tag: widget.tag);
      }
    }
    remove?.call();
  }

  void _subscribeToController() {
    remove?.call();
    if (widget.id == null) {
      remove = controller?.addListener(getUpdate);
    } else {
      remove = controller?.addListenerWithId(widget.id, getUpdate);
    }
  }

  @override
  Widget build(BuildContext context) => widget.builder(controller);
}
